@@ -14,19 +14,18 @@ import (
 	"go/token"
 	"go/types"
 	"slices"
-	"sort"
 	"strconv"
 	"strings"
-	"text/scanner"
 
 	"golang.org/x/tools/go/analysis"
-	"golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/edge"
 	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/parsego"
-	goplsastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
+	"golang.org/x/tools/gopls/internal/util/cursorutil"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/typesinternal"
 )
 
@@ -55,16 +54,16 @@ func extractVariable(pkg *cache.Package, pgf *parsego.File, start, end token.Pos
 		info = pkg.TypesInfo()
 		file = pgf.File
 	)
-	// TODO(adonovan): simplify, using Cursor.
-	exprs, err := canExtractVariable(info, pgf.Cursor, start, end, all)
+	curExprs, err := canExtractVariable(info, pgf.Cursor, start, end, all)
 	if err != nil {
 		return nil, nil, fmt.Errorf("cannot extract: %v", err)
 	}
+	expr0 := curExprs[0].Node().(ast.Expr)
 
 	// innermost scope enclosing ith expression
-	exprScopes := make([]*types.Scope, len(exprs))
-	for i, e := range exprs {
-		exprScopes[i] = info.Scopes[file].Innermost(e.Pos())
+	exprScopes := make([]*types.Scope, len(curExprs))
+	for i, curExpr := range curExprs {
+		exprScopes[i] = info.Scopes[file].Innermost(curExpr.Node().Pos())
 	}
 
 	hasCollision := func(name string) bool {
@@ -75,12 +74,12 @@ func extractVariable(pkg *cache.Package, pgf *parsego.File, start, end token.Pos
 		}
 		return false
 	}
-	constant := info.Types[exprs[0]].Value != nil
+	constant := info.Types[expr0].Value != nil
 
 	// Generate name(s) for new declaration.
 	baseName := cond(constant, "newConst", "newVar")
 	var lhsNames []string
-	switch expr := exprs[0].(type) {
+	switch expr := expr0.(type) {
 	case *ast.CallExpr:
 		tup, ok := info.TypeOf(expr).(*types.Tuple)
 		if !ok {
@@ -122,7 +121,7 @@ Outer:
 		}
 	}
 
-	var visiblePath []ast.Node
+	visible := curExprs[0] // Insert newVar inside commonScope before the first occurrence of the expression.
 	if commonScope != exprScopes[0] {
 		// This means the first expr within function body is not the largest scope,
 		// we need to find the scope immediately follow the common
@@ -134,12 +133,9 @@ Outer:
 			}
 			child = p
 		}
-		visiblePath, _ = astutil.PathEnclosingInterval(file, child.Pos(), child.End())
-	} else {
-		// Insert newVar inside commonScope before the first occurrence of the expression.
-		visiblePath, _ = astutil.PathEnclosingInterval(file, exprs[0].Pos(), exprs[0].End())
+		visible, _ = pgf.Cursor.FindByPos(child.Pos(), child.End())
 	}
-	variables, err := collectFreeVars(info, file, exprs[0].Pos(), exprs[0].End(), exprs[0])
+	variables, err := collectFreeVars(info, file, expr0.Pos(), expr0.End(), expr0)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -159,17 +155,19 @@ Outer:
 		indentation string
 		stmtOK      bool // ok to use ":=" instead of var/const decl?
 	)
-	if funcDecl, ok := visiblePath[len(visiblePath)-2].(*ast.FuncDecl); ok && goplsastutil.NodeContains(funcDecl.Body, start) {
-		before, err := stmtToInsertVarBefore(visiblePath, variables)
+	if funcDecl, _ := cursorutil.FirstEnclosing[*ast.FuncDecl](visible); funcDecl != nil &&
+		astutil.NodeContainsPos(funcDecl.Body, start) {
+
+		beforePos, err := stmtToInsertVarBefore(visible, variables)
 		if err != nil {
 			return nil, nil, fmt.Errorf("cannot find location to insert extraction: %v", err)
 		}
 		// Within function: compute appropriate statement indentation.
-		indent, err := pgf.Indentation(before.Pos())
+		indent, err := pgf.Indentation(beforePos)
 		if err != nil {
 			return nil, nil, err
 		}
-		insertPos = before.Pos()
+		insertPos = beforePos
 		indentation = "\n" + indent
 
 		// Currently, we always extract a constant expression
@@ -185,9 +183,10 @@ Outer:
 		stmtOK = !constant
 	} else {
 		// Outside any statement: insert before the current
-		// declaration, without indentation.
-		currentDecl := visiblePath[len(visiblePath)-2]
-		insertPos = currentDecl.Pos()
+		// top-level declaration, without indentation.
+		for curDecl := range visible.Enclosing((*ast.FuncDecl)(nil), (*ast.GenDecl)(nil)) {
+			insertPos = curDecl.Node().Pos()
+		}
 		indentation = "\n"
 	}
 
@@ -222,7 +221,7 @@ Outer:
 			Specs: []ast.Spec{
 				&ast.ValueSpec{
 					Names:  names,
-					Values: []ast.Expr{exprs[0]},
+					Values: []ast.Expr{expr0},
 				},
 			},
 		}
@@ -236,7 +235,7 @@ Outer:
 		newNode = &ast.AssignStmt{
 			Tok: token.DEFINE,
 			Lhs: lhs,
-			Rhs: []ast.Expr{exprs[0]},
+			Rhs: []ast.Expr{expr0},
 		}
 	}
 
@@ -252,7 +251,8 @@ Outer:
 		End:     insertPos,
 		NewText: []byte(assignment),
 	}}
-	for _, e := range exprs {
+	for _, curExpr := range curExprs {
+		e := curExpr.Node()
 		textEdits = append(textEdits, analysis.TextEdit{
 			Pos:     e.Pos(),
 			End:     e.End(),
@@ -293,18 +293,21 @@ Outer:
 //
 // `x` is a free variable defined in the IfStmt, we should not insert
 // the extracted expression outside the IfStmt scope, instead, return an error.
-func stmtToInsertVarBefore(path []ast.Node, variables []*variable) (ast.Stmt, error) {
-	enclosingIndex := -1 // index in path of enclosing stmt
-	for i, p := range path {
-		if _, ok := p.(ast.Stmt); ok {
-			enclosingIndex = i
-			break
+func stmtToInsertVarBefore(cur inspector.Cursor, variables []*variable) (token.Pos, error) {
+	// Walk up to enclosing statement.
+	{
+		var curStmt inspector.Cursor
+		for cur := range cur.Enclosing() {
+			if is[ast.Stmt](cur.Node()) {
+				curStmt = cur
+				break
+			}
 		}
+		if !astutil.CursorValid(curStmt) {
+			return 0, fmt.Errorf("no enclosing statement")
+		}
+		cur = curStmt
 	}
-	if enclosingIndex == -1 {
-		return nil, fmt.Errorf("no enclosing statement")
-	}
-	enclosingStmt := path[enclosingIndex].(ast.Stmt)
 
 	// hasFreeVar reports if any free variables is defined inside stmt (which may be nil).
 	// If true, indicates that the insertion point will sit before the variable declaration.
@@ -313,7 +316,7 @@ func stmtToInsertVarBefore(path []ast.Node, variables []*variable) (ast.Stmt, er
 			return false
 		}
 		for _, v := range variables {
-			if goplsastutil.NodeContains(stmt, v.obj.Pos()) {
+			if astutil.NodeContainsPos(stmt, v.obj.Pos()) {
 				return true
 			}
 		}
@@ -322,119 +325,114 @@ func stmtToInsertVarBefore(path []ast.Node, variables []*variable) (ast.Stmt, er
 
 	// baseIfStmt walks up the if/else-if chain until we get to
 	// the top of the current if chain.
-	baseIfStmt := func(index int) (ast.Stmt, error) {
-		stmt := path[index]
-		for _, node := range path[index+1:] {
-			ifStmt, ok := node.(*ast.IfStmt)
-			if !ok || ifStmt.Else != stmt {
-				break
+	baseIfStmt := func(curIf inspector.Cursor) (token.Pos, error) {
+		for astutil.IsChildOf(curIf, edge.IfStmt_Else) {
+			curIf = curIf.Parent()
+			if hasFreeVar(curIf.Node().(*ast.IfStmt).Init) {
+				return 0, fmt.Errorf("else-if's init has free variable")
 			}
-			if hasFreeVar(ifStmt.Init) {
-				return nil, fmt.Errorf("Else's init statement has free variable declaration")
-			}
-			stmt = ifStmt
 		}
-		return stmt.(ast.Stmt), nil
+		return curIf.Node().Pos(), nil
 	}
 
-	switch enclosingStmt := enclosingStmt.(type) {
+	stmt := cur.Node()
+	switch stmt := stmt.(type) {
 	case *ast.IfStmt:
-		if hasFreeVar(enclosingStmt.Init) {
-			return nil, fmt.Errorf("IfStmt's init statement has free variable declaration")
+		if hasFreeVar(stmt.Init) {
+			return 0, fmt.Errorf("if statement's init has free variable")
 		}
-		// The enclosingStmt is inside of the if declaration,
+		// stmt is inside of the if declaration.
 		// We need to check if we are in an else-if stmt and
 		// get the base if statement.
-		return baseIfStmt(enclosingIndex)
+		return baseIfStmt(cur)
+
 	case *ast.CaseClause:
-		// Get the enclosing switch stmt if the enclosingStmt is
-		// inside of the case statement.
-		for _, node := range path[enclosingIndex+1:] {
-			switch stmt := node.(type) {
+		for curSwitch := range cur.Enclosing((*ast.SwitchStmt)(nil), (*ast.TypeSwitchStmt)(nil)) {
+			swtch := curSwitch.Node()
+			var init ast.Stmt
+			switch swtch := swtch.(type) {
 			case *ast.SwitchStmt:
-				if hasFreeVar(stmt.Init) {
-					return nil, fmt.Errorf("SwitchStmt's init statement has free variable declaration")
-				}
-				return stmt, nil
+				init = swtch.Init
 			case *ast.TypeSwitchStmt:
-				if hasFreeVar(stmt.Init) {
-					return nil, fmt.Errorf("TypeSwitchStmt's init statement has free variable declaration")
-				}
-				return stmt, nil
+				init = swtch.Init
 			}
+			if hasFreeVar(init) {
+				return 0, fmt.Errorf("switch's init has free variable")
+			}
+			return swtch.Pos(), nil
 		}
 	}
+
 	// Check if the enclosing statement is inside another node.
-	switch parent := path[enclosingIndex+1].(type) {
+	switch parent := cur.Parent().Node().(type) {
 	case *ast.IfStmt:
 		if hasFreeVar(parent.Init) {
-			return nil, fmt.Errorf("IfStmt's init statement has free variable declaration")
+			return 0, fmt.Errorf("if-statement's init has free variable")
 		}
-		return baseIfStmt(enclosingIndex + 1)
+		return baseIfStmt(cur.Parent())
+
 	case *ast.ForStmt:
-		if parent.Init == enclosingStmt || parent.Post == enclosingStmt {
-			return parent, nil
+		switch cur.Node() {
+		case parent.Init, parent.Post:
+			return parent.Pos(), nil
 		}
+
 	case *ast.SwitchStmt:
 		if hasFreeVar(parent.Init) {
-			return nil, fmt.Errorf("SwitchStmt's init statement has free variable declaration")
+			return 0, fmt.Errorf("switch's init has free variable")
 		}
-		return parent, nil
+		return parent.Pos(), nil
+
 	case *ast.TypeSwitchStmt:
 		if hasFreeVar(parent.Init) {
-			return nil, fmt.Errorf("TypeSwitchStmt's init statement has free variable declaration")
+			return 0, fmt.Errorf("switch's init has free variable")
 		}
-		return parent, nil
+		return parent.Pos(), nil
 	}
-	return enclosingStmt, nil
+	return stmt.Pos(), nil
 }
 
-// canExtractVariable reports whether the code in the given range can be
-// extracted to a variable (or constant). It returns the selected expression or, if 'all',
-// all structurally equivalent expressions within the same function body, in lexical order.
-func canExtractVariable(info *types.Info, curFile inspector.Cursor, start, end token.Pos, all bool) ([]ast.Expr, error) {
+// canExtractVariable reports whether the code in the given range can
+// be extracted to a variable (or constant). It returns (cursors for)
+// the selected expression or, if 'all', all structurally equivalent
+// expressions within the same function body, in lexical order.
+func canExtractVariable(info *types.Info, curFile inspector.Cursor, start, end token.Pos, all bool) ([]inspector.Cursor, error) {
 	if start == end {
 		return nil, fmt.Errorf("empty selection")
 	}
-	file := curFile.Node().(*ast.File)
-	// TODO(adonovan): simplify, using Cursor.
-	path, exact := astutil.PathEnclosingInterval(file, start, end)
-	if !exact {
+
+	_, curStart, curEnd, err := astutil.Select(curFile, start, end)
+	if err != nil {
+		return nil, err
+	}
+	expr, ok := curStart.Node().(ast.Expr)
+	if !ok || curEnd != curStart {
 		return nil, fmt.Errorf("selection is not an expression")
 	}
-	if len(path) == 0 {
-		return nil, bug.Errorf("no path enclosing interval")
-	}
-	for _, n := range path {
-		if _, ok := n.(*ast.ImportSpec); ok {
-			return nil, fmt.Errorf("cannot extract variable or constant in an import block")
-		}
-	}
-	expr, ok := path[0].(ast.Expr)
-	if !ok {
-		return nil, fmt.Errorf("selection is not an expression") // e.g. statement
+	if imp, _ := cursorutil.FirstEnclosing[*ast.ImportSpec](curStart); imp != nil {
+		return nil, fmt.Errorf("cannot extract variable or constant in an import block")
 	}
 	if tv, ok := info.Types[expr]; !ok || !tv.IsValue() || tv.Type == nil || tv.HasOk() {
 		// e.g. type, builtin, x.(type), 2-valued m[k], or ill-typed
 		return nil, fmt.Errorf("selection is not a single-valued expression")
 	}
 
-	var exprs []ast.Expr
+	var curExprs []inspector.Cursor
 	if !all {
-		exprs = append(exprs, expr)
-	} else if funcDecl, ok := path[len(path)-2].(*ast.FuncDecl); ok {
+		curExprs = append(curExprs, curStart)
+	} else if funcDecl, curFuncDecl := cursorutil.FirstEnclosing[*ast.FuncDecl](curStart); funcDecl != nil && funcDecl.Body != nil {
 		// Find all expressions in the same function body that
 		// are equal to the selected expression.
-		ast.Inspect(funcDecl.Body, func(n ast.Node) bool {
-			if e, ok := n.(ast.Expr); ok {
-				if goplsastutil.Equal(e, expr, func(x, y *ast.Ident) bool {
+		for cur := range curFuncDecl.ChildAt(edge.FuncDecl_Body, -1).Preorder() {
+			if e, ok := cur.Node().(ast.Expr); ok {
+				if astutil.Equal(e, expr, func(x, y *ast.Ident) bool {
 					xobj, yobj := info.ObjectOf(x), info.ObjectOf(y)
 					// The two identifiers must resolve to the same object,
 					// or to a declaration within the candidate expression.
 					// (This allows two copies of "func (x int) { print(x) }"
 					// to match.)
-					if xobj != nil && goplsastutil.NodeContains(e, xobj.Pos()) &&
-						yobj != nil && goplsastutil.NodeContains(expr, yobj.Pos()) {
+					if xobj != nil && astutil.NodeContainsPos(e, xobj.Pos()) &&
+						yobj != nil && astutil.NodeContainsPos(expr, yobj.Pos()) {
 						return x.Name == y.Name
 					}
 					// Use info.Uses to avoid including declaration, for example,
@@ -446,11 +444,10 @@ func canExtractVariable(info *types.Info, curFile inspector.Cursor, start, end t
 					xuse := info.Uses[x]
 					return xuse != nil && xuse == info.Uses[y]
 				}) {
-					exprs = append(exprs, e)
+					curExprs = append(curExprs, cur)
 				}
 			}
-			return true
-		})
+		}
 	} else {
 		return nil, fmt.Errorf("node %T is not inside a function", expr)
 	}
@@ -485,26 +482,16 @@ func canExtractVariable(info *types.Info, curFile inspector.Cursor, start, end t
 	//   newVar := &f.bar
 	//   x := *newVar + 1
 	//   *newVar = 2
-	for _, e := range exprs {
-		path, _ := astutil.PathEnclosingInterval(file, e.Pos(), e.End())
-		for _, n := range path {
-			if assignment, ok := n.(*ast.AssignStmt); ok {
-				if slices.Contains(assignment.Lhs, e) {
-					return nil, fmt.Errorf("node %T is in LHS of an AssignStmt", expr)
-				}
-				break
-			}
-			if value, ok := n.(*ast.ValueSpec); ok {
-				for _, name := range value.Names {
-					if name == e {
-						return nil, fmt.Errorf("node %T is in LHS of a ValueSpec", expr)
-					}
-				}
-				break
-			}
+	for _, curExpr := range curExprs {
+		switch ek, _ := curExpr.ParentEdge(); ek {
+		case edge.AssignStmt_Lhs:
+			return nil, fmt.Errorf("node %T is in LHS of an AssignStmt", expr)
+		case edge.ValueSpec_Names:
+			return nil, fmt.Errorf("node %T is in LHS of a ValueSpec", expr)
 		}
 	}
-	return exprs, nil
+
+	return curExprs, nil
 }
 
 // freshName returns an identifier based on prefix (perhaps with a
@@ -594,6 +581,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		pkg  = cpkg.Types()
 		info = cpkg.TypesInfo()
 		src  = pgf.Src
+		file = pgf.File
 	)
 
 	errorPrefix := "extractFunction"
@@ -601,31 +589,23 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		errorPrefix = "extractMethod"
 	}
 
-	file := pgf.Cursor.Node().(*ast.File)
-	// TODO(adonovan): simplify, using Cursor.
-	tok := fset.File(file.FileStart)
-	if tok == nil {
-		return nil, nil, bug.Errorf("no file for position")
-	}
-	p, ok, methodOk, err := canExtractFunction(tok, start, end, src, pgf.Cursor)
+	p, ok, methodOk, err := canExtractFunction(pgf.Cursor, start, end)
 	if (!ok && !isMethod) || (!methodOk && isMethod) {
 		return nil, nil, fmt.Errorf("%s: cannot extract %s: %v", errorPrefix,
 			safetoken.StartPosition(fset, start), err)
 	}
-	tok, path, start, end, outer, node := p.tok, p.path, p.start, p.end, p.outer, p.node
+	curEnclosing, curStart, curEnd, curFuncDecl := p.curEnclosing, p.curStart, p.curEnd, p.curFuncDecl
+
+	// Narrow (start, end) to the located nodes.
+	start, end = curStart.Node().Pos(), curEnd.Node().End()
+
+	outer := curFuncDecl.Node().(*ast.FuncDecl)
 
 	// A return statement is non-nested if its parent node is equal to the parent node
 	// of the first node in the selection. These cases must be handled separately because
 	// non-nested return statements are guaranteed to execute.
 	var hasNonNestedReturn bool
-	curStart, ok := pgf.Cursor.FindNode(node)
-	if !ok {
-		return nil, nil, bug.Errorf("cannot find Cursor for start Node")
-	}
-	curOuter, ok := pgf.Cursor.FindNode(outer)
-	if !ok {
-		return nil, nil, bug.Errorf("cannot find Cursor for start Node")
-	}
+
 	// Determine whether all return statements in the selection are
 	// error-handling return statements. They must be of the form:
 	// if err != nil {
@@ -634,10 +614,12 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 	// If all return statements in the extracted block have a non-nil error, we
 	// can replace the "shouldReturn" check with an error check to produce a
 	// more concise output.
-	allReturnsFinalErr := true // all ReturnStmts have final 'err' expression
-	hasReturn := false         // selection contains a ReturnStmt
-	filter := []ast.Node{(*ast.ReturnStmt)(nil), (*ast.FuncLit)(nil)}
-	curOuter.Inspect(filter, func(cur inspector.Cursor) (descend bool) {
+	var (
+		allReturnsFinalErr = true  // all ReturnStmts have final 'err' expression
+		hasReturn          = false // selection contains a ReturnStmt
+		filter             = []ast.Node{(*ast.ReturnStmt)(nil), (*ast.FuncLit)(nil)}
+	)
+	curEnclosing.Inspect(filter, func(cur inspector.Cursor) (descend bool) {
 		if funcLit, ok := cur.Node().(*ast.FuncLit); ok {
 			// Exclude return statements in function literals because they don't affect the refactor.
 			// Keep descending into func lits whose declaration is not included in the extracted block.
@@ -691,7 +673,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 	// we must determine the signature of the extracted function. We will then replace
 	// the block with an assignment statement that calls the extracted function with
 	// the appropriate parameters and return values.
-	variables, err := collectFreeVars(info, file, start, end, path[0])
+	variables, err := collectFreeVars(info, file, start, end, curEnclosing.Node())
 	if err != nil {
 		return nil, nil, err
 	}
@@ -703,7 +685,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		receiverObj  types.Object
 	)
 	if isMethod {
-		if outer == nil || outer.Recv == nil || len(outer.Recv.List) == 0 {
+		if outer.Recv == nil || len(outer.Recv.List) == 0 {
 			return nil, nil, fmt.Errorf("%s: cannot extract need method receiver", errorPrefix)
 		}
 		receiver = outer.Recv.List[0]
@@ -772,7 +754,9 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		// cannot be its own reassignment or redefinition (objOverriden).
 		vscope := v.obj.Parent()
 		if vscope == nil {
-			return nil, nil, fmt.Errorf("parent nil")
+			// v.obj could be a field on an anonymous struct. We'll examine the
+			// struct in a different iteration so don't return an error here.
+			continue
 		}
 		isUsed, firstUseAfter := objUsed(info, end, vscope.End(), v.obj)
 		if v.assigned && isUsed && !varOverridden(info, firstUseAfter, v.obj, v.free, outer) {
@@ -831,19 +815,13 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 	// the top-level declaration. We inspect the top-level declaration to look for variables
 	// as well as for code replacement.
 	enclosing := outer.Type
-	for _, p := range path {
-		if p == enclosing {
-			break
-		}
-		if fl, ok := p.(*ast.FuncLit); ok {
-			enclosing = fl.Type
-			break
-		}
+	if funcLit, _ := cursorutil.FirstEnclosing[*ast.FuncLit](curEnclosing); funcLit != nil {
+		enclosing = funcLit.Type
 	}
 
 	// We put the selection in a constructed file. We can then traverse and edit
 	// the extracted selection without modifying the original AST.
-	startOffset, endOffset, err := safetoken.Offsets(tok, start, end)
+	startOffset, endOffset, err := safetoken.Offsets(pgf.Tok, start, end)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -955,7 +933,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		// statements in the selection. Update the type signature of the extracted
 		// function and construct the if statement that will be inserted in the enclosing
 		// function.
-		retVars, ifReturn, err = generateReturnInfo(enclosing, pkg, path, file, info, start, end, hasNonNestedReturn, isErrHandlingReturnsCase)
+		retVars, ifReturn, err = generateReturnInfo(enclosing, pkg, curEnclosing.Node().Pos(), file, info, start, end, hasNonNestedReturn, isErrHandlingReturnsCase)
 		if err != nil {
 			return nil, nil, err
 		}
@@ -1177,13 +1155,13 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 
 	// We're going to replace the whole enclosing function,
 	// so preserve the text before and after the selected block.
-	outerStart, outerEnd, err := safetoken.Offsets(tok, outer.Pos(), outer.End())
+	outerStart, outerEnd, err := pgf.NodeOffsets(outer)
 	if err != nil {
 		return nil, nil, err
 	}
 	before := src[outerStart:startOffset]
 	after := src[endOffset:outerEnd]
-	indent, err := pgf.Indentation(node.Pos())
+	indent, err := pgf.Indentation(start)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -1212,7 +1190,7 @@ func extractFunctionMethod(cpkg *cache.Package, pgf *parsego.File, start, end to
 		fmt.Fprintf(&fullReplacement, "%[1]sswitch %[2]s {%[1]s", newLineIndent, ctrlVar)
 		for i, br := range freeBranches {
 			// Preserve spacing at the beginning of the line containing the branch statement.
-			startPos := tok.LineStart(safetoken.Line(tok, br.Pos()))
+			startPos := pgf.Tok.LineStart(safetoken.Line(pgf.Tok, br.Pos()))
 			text, err := pgf.PosText(startPos, br.End())
 			if err != nil {
 				return nil, nil, err
@@ -1272,88 +1250,6 @@ func moveParamToFrontIfFound(params []ast.Expr, paramTypes []*ast.Field, x, sel
 	}
 }
 
-// adjustRangeForCommentsAndWhiteSpace adjusts the given range to exclude unnecessary leading or
-// trailing whitespace characters from selection as well as leading or trailing comments.
-// In the following example, each line of the if statement is indented once. There are also two
-// extra spaces after the sclosing bracket before the line break and a comment.
-//
-// \tif (true) {
-// \t    _ = 1
-// \t} // hello \n
-//
-// By default, a valid range begins at 'if' and ends at the first whitespace character
-// after the '}'. But, users are likely to highlight full lines rather than adjusting
-// their cursors for whitespace. To support this use case, we must manually adjust the
-// ranges to match the correct AST node. In this particular example, we would adjust
-// rng.Start forward to the start of 'if' and rng.End backward to after '}'.
-func adjustRangeForCommentsAndWhiteSpace(tok *token.File, start, end token.Pos, content []byte, curFile inspector.Cursor) (token.Pos, token.Pos, error) {
-	file := curFile.Node().(*ast.File)
-	// TODO(adonovan): simplify, using Cursor.
-
-	// Adjust the end of the range to after leading whitespace and comments.
-	prevStart := token.NoPos
-	startComment := sort.Search(len(file.Comments), func(i int) bool {
-		// Find the index for the first comment that ends after range start.
-		return file.Comments[i].End() > start
-	})
-	for prevStart != start {
-		prevStart = start
-		// If start is within a comment, move start to the end
-		// of the comment group.
-		if startComment < len(file.Comments) && file.Comments[startComment].Pos() <= start && start < file.Comments[startComment].End() {
-			start = file.Comments[startComment].End()
-			startComment++
-		}
-		// Move forwards to find a non-whitespace character.
-		offset, err := safetoken.Offset(tok, start)
-		if err != nil {
-			return 0, 0, err
-		}
-		for offset < len(content) && isGoWhiteSpace(content[offset]) {
-			offset++
-		}
-		start = tok.Pos(offset)
-	}
-
-	// Adjust the end of the range to before trailing whitespace and comments.
-	prevEnd := token.NoPos
-	endComment := sort.Search(len(file.Comments), func(i int) bool {
-		// Find the index for the first comment that ends after the range end.
-		return file.Comments[i].End() >= end
-	})
-	// Search will return n if not found, so we need to adjust if there are no
-	// comments that would match.
-	if endComment == len(file.Comments) {
-		endComment = -1
-	}
-	for prevEnd != end {
-		prevEnd = end
-		// If end is within a comment, move end to the start
-		// of the comment group.
-		if endComment >= 0 && file.Comments[endComment].Pos() < end && end <= file.Comments[endComment].End() {
-			end = file.Comments[endComment].Pos()
-			endComment--
-		}
-		// Move backwards to find a non-whitespace character.
-		offset, err := safetoken.Offset(tok, end)
-		if err != nil {
-			return 0, 0, err
-		}
-		for offset > 0 && isGoWhiteSpace(content[offset-1]) {
-			offset--
-		}
-		end = tok.Pos(offset)
-	}
-
-	return start, end, nil
-}
-
-// isGoWhiteSpace returns true if b is a considered white space in
-// Go as defined by scanner.GoWhitespace.
-func isGoWhiteSpace(b byte) bool {
-	return uint64(scanner.GoWhitespace)&(1<<uint(b)) != 0
-}
-
 // variable describes the status of a variable within a selection.
 type variable struct {
 	obj types.Object
@@ -1566,6 +1462,12 @@ func collectFreeVars(info *types.Info, file *ast.File, start, end token.Pos, nod
 		if !ok {
 			return nil, fmt.Errorf("no seen types.Object for %v", obj)
 		}
+		if named, ok := v.obj.Type().(typesinternal.NamedOrAlias); ok {
+			namedPos := named.Obj().Pos()
+			if isLocal(named.Obj()) && !(start <= namedPos && namedPos <= end) {
+				return nil, fmt.Errorf("Cannot extract selection: the code refers to a local type whose definition lies outside the extracted block")
+			}
+		}
 		variables = append(variables, v)
 	}
 	return variables, nil
@@ -1593,86 +1495,55 @@ func referencesObj(info *types.Info, expr ast.Expr, obj types.Object) bool {
 }
 
 type fnExtractParams struct {
-	tok        *token.File
-	start, end token.Pos
-	path       []ast.Node
-	outer      *ast.FuncDecl
-	node       ast.Node
+	curStart, curEnd inspector.Cursor // first and last nodes wholly enclosed by selection
+	curEnclosing     inspector.Cursor // node that encloses selection (e.g. a BlockStmt)
+	curFuncDecl      inspector.Cursor // enclosing *ast.FuncDecl
 }
 
 // canExtractFunction reports whether the code in the given range can be
 // extracted to a function.
-func canExtractFunction(tok *token.File, start, end token.Pos, src []byte, curFile inspector.Cursor) (*fnExtractParams, bool, bool, error) {
+func canExtractFunction(curFile inspector.Cursor, start, end token.Pos) (*fnExtractParams, bool, bool, error) {
 	if start == end {
 		return nil, false, false, fmt.Errorf("start and end are equal")
 	}
-	var err error
-	file := curFile.Node().(*ast.File)
-	// TODO(adonovan): simplify, using Cursor.
-	start, end, err = adjustRangeForCommentsAndWhiteSpace(tok, start, end, src, curFile)
+
+	curEnclosing, curStart, curEnd, err := astutil.Select(curFile, start, end)
 	if err != nil {
 		return nil, false, false, err
 	}
-	path, _ := astutil.PathEnclosingInterval(file, start, end)
-	if len(path) == 0 {
-		return nil, false, false, fmt.Errorf("no path enclosing interval")
-	}
+
 	// Node that encloses the selection must be a statement.
 	// TODO: Support function extraction for an expression.
-	_, ok := path[0].(ast.Stmt)
-	if !ok {
+	if !is[ast.Stmt](curEnclosing.Node()) {
 		return nil, false, false, fmt.Errorf("node is not a statement")
 	}
 
 	// Find the function declaration that encloses the selection.
-	var outer *ast.FuncDecl
-	for _, p := range path {
-		if p, ok := p.(*ast.FuncDecl); ok {
-			outer = p
-			break
-		}
-	}
-	if outer == nil {
+	funcDecl, curFuncDecl := cursorutil.FirstEnclosing[*ast.FuncDecl](curEnclosing)
+	if funcDecl == nil {
 		return nil, false, false, fmt.Errorf("no enclosing function")
 	}
 
-	// Find the nodes at the start and end of the selection.
-	var startNode, endNode ast.Node
-	ast.Inspect(outer, func(n ast.Node) bool {
-		if n == nil {
-			return false
-		}
-		// Do not override 'start' with a node that begins at the same location
-		// but is nested further from 'outer'.
-		if startNode == nil && n.Pos() == start && n.End() <= end {
-			startNode = n
-		}
-		if endNode == nil && n.End() == end && n.Pos() >= start {
-			endNode = n
-		}
-		return n.Pos() <= end
-	})
-	if startNode == nil || endNode == nil {
-		return nil, false, false, fmt.Errorf("range does not map to AST nodes")
-	}
-	// If the region is a blockStmt, use the first and last nodes in the block
-	// statement.
-	// <rng.start>{ ... }<rng.end> => { <rng.start>...<rng.end> }
-	if blockStmt, ok := startNode.(*ast.BlockStmt); ok {
-		if len(blockStmt.List) == 0 {
+	// If the selection is a block statement, use its first and last statements.
+	// «{ ... }»  =>  { «...» }
+	if is[*ast.BlockStmt](curStart.Node()) && curStart == curEnd {
+		var (
+			first, ok1 = curStart.FirstChild()
+			last, ok2  = curStart.LastChild()
+		)
+		if !(ok1 && ok2) {
 			return nil, false, false, fmt.Errorf("range maps to empty block statement")
 		}
-		startNode, endNode = blockStmt.List[0], blockStmt.List[len(blockStmt.List)-1]
-		start, end = startNode.Pos(), endNode.End()
+		curStart = first
+		curEnd = last
 	}
+
 	return &fnExtractParams{
-		tok:   tok,
-		start: start,
-		end:   end,
-		path:  path,
-		outer: outer,
-		node:  startNode,
-	}, true, outer.Recv != nil, nil
+		curStart:     curStart,
+		curEnd:       curEnd,
+		curEnclosing: curEnclosing,
+		curFuncDecl:  curFuncDecl,
+	}, true, funcDecl.Recv != nil, nil
 }
 
 // objUsed checks if the object is used within the range. It returns the first
@@ -1768,7 +1639,7 @@ func parseStmts(fset *token.FileSet, src []byte) (*ast.BlockStmt, []*ast.Comment
 // signature of the extracted function. We prepare names, signatures, and "zero values" that
 // represent the new variables. We also use this information to construct the if statement that
 // is inserted below the call to the extracted function.
-func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, path []ast.Node, file *ast.File, info *types.Info, start, end token.Pos, hasNonNestedReturns bool, isErrHandlingReturnsCase bool) ([]*returnVariable, *ast.IfStmt, error) {
+func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, at token.Pos, file *ast.File, info *types.Info, start, end token.Pos, hasNonNestedReturns bool, isErrHandlingReturnsCase bool) ([]*returnVariable, *ast.IfStmt, error) {
 	var retVars []*returnVariable
 	var cond *ast.Ident
 	// Generate information for the values in the return signature of the enclosing function.
@@ -1795,7 +1666,7 @@ func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, path []ast.
 				} else if n, ok := varNameForType(typ); ok {
 					bestName = n
 				}
-				retName, idx := freshNameOutsideRange(info, file, path[0].Pos(), start, end, bestName, nameIdx[bestName])
+				retName, idx := freshNameOutsideRange(info, file, at, start, end, bestName, nameIdx[bestName])
 				nameIdx[bestName] = idx
 				z, isValid := typesinternal.ZeroExpr(typ, qual)
 				if !isValid {
@@ -1814,7 +1685,7 @@ func generateReturnInfo(enclosing *ast.FuncType, pkg *types.Package, path []ast.
 		results := getNames(retVars)
 		if !isErrHandlingReturnsCase {
 			// Generate information for the added bool value.
-			name, _ := freshNameOutsideRange(info, file, path[0].Pos(), start, end, "shouldReturn", 0)
+			name, _ := freshNameOutsideRange(info, file, at, start, end, "shouldReturn", 0)
 			cond = &ast.Ident{Name: name}
 			retVars = append(retVars, &returnVariable{
 				name:    cond,
@@ -1893,9 +1764,11 @@ func adjustReturnStatements(returnTypes []*ast.Field, seenVars map[types.Object]
 	// extracted function. We set the bool to 'true' because, if these return statements
 	// execute, the extracted function terminates early, and the enclosing function must
 	// return as well.
+	var shouldReturnCond []ast.Expr
 	if !isErrHandlingReturnsCase {
-		zeroVals = append(zeroVals, ast.NewIdent("true"))
+		shouldReturnCond = append(shouldReturnCond, ast.NewIdent("true"))
 	}
+
 	ast.Inspect(extractedBlock, func(n ast.Node) bool {
 		if n == nil {
 			return false
@@ -1905,7 +1778,7 @@ func adjustReturnStatements(returnTypes []*ast.Field, seenVars map[types.Object]
 			return false
 		}
 		if n, ok := n.(*ast.ReturnStmt); ok {
-			n.Results = slices.Concat(zeroVals, n.Results)
+			n.Results = slices.Concat(zeroVals, n.Results, shouldReturnCond)
 			return false
 		}
 		return true
