@@ -52,7 +52,7 @@ import (
 	"go/token"
 	"go/types"
 	"maps"
-	"path"
+	"os"
 	"path/filepath"
 	"regexp"
 	"slices"
@@ -61,7 +61,8 @@ import (
 	"strings"
 
 	"golang.org/x/mod/modfile"
-	"golang.org/x/tools/go/ast/astutil"
+	goastutil "golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/go/types/objectpath"
 	"golang.org/x/tools/go/types/typeutil"
 	"golang.org/x/tools/gopls/internal/cache"
@@ -69,11 +70,11 @@ import (
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/protocol"
-	goplsastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
-	"golang.org/x/tools/gopls/internal/util/moreiters"
+	"golang.org/x/tools/gopls/internal/util/cursorutil"
+	"golang.org/x/tools/gopls/internal/util/pathutil"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	internalastutil "golang.org/x/tools/internal/astutil"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/diff"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/typesinternal"
@@ -134,23 +135,38 @@ func PrepareRename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle,
 	if err != nil {
 		return nil, nil, err
 	}
+	cur, _ := pgf.Cursor.FindByPos(pos, pos) // can't fail
 
 	// Check if we're in a 'func' keyword. If so, we hijack the renaming to
 	// change the function signature.
-	if item, err := prepareRenameFuncSignature(pgf, pos); err != nil {
+	if item, err := prepareRenameFuncSignature(pgf, pos, cur); err != nil {
 		return nil, nil, err
 	} else if item != nil {
 		return item, nil, nil
 	}
 
-	targets, node, err := objectsAt(pkg.TypesInfo(), pgf.File, pos)
+	targets, err := objectsAt(pkg.TypesInfo(), cur)
 	if err != nil {
-		return nil, nil, err
-	}
-	var obj types.Object
-	for obj = range targets {
-		break // pick one arbitrarily
+		// Check if we are renaming an ident inside its doc comment. The call to
+		// objectsAt will have returned an error in this case.
+		// TODO(adonovan): move this logic into objectsAt.
+		var ok bool
+		cur, ok = docCommentPosToIdent(pgf, pos, cur)
+		if !ok {
+			return nil, nil, err
+		}
+		id := cur.Node().(*ast.Ident)
+		obj := pkg.TypesInfo().Defs[id]
+		if obj == nil {
+			return nil, nil, fmt.Errorf("error fetching Object for ident %q", id.Name)
+		}
+		// Change rename target to the ident.
+		targets = []objectAt{{obj, cur}}
 	}
+
+	// Pick a representative object arbitrarily.
+	// (All share the same name, pos, and kind.)
+	obj, node := targets[0].obj, targets[0].cur.Node()
 	if err := checkRenamable(obj, node); err != nil {
 		return nil, nil, err
 	}
@@ -182,12 +198,14 @@ func prepareRenamePackageName(ctx context.Context, snapshot *cache.Snapshot, pgf
 	if meta.Name == "main" {
 		return nil, fmt.Errorf("can't rename package \"main\"")
 	}
+	// TODO(mkalil): support renaming x_test packages.
 	if strings.HasSuffix(string(meta.Name), "_test") {
 		return nil, fmt.Errorf("can't rename x_test packages")
 	}
 	if meta.Module == nil {
 		return nil, fmt.Errorf("can't rename package: missing module information for package %q", meta.PkgPath)
 	}
+	// TODO(mkalil): why is renaming the root package of a module unsupported?
 	if meta.Module.Path == string(meta.PkgPath) {
 		return nil, fmt.Errorf("can't rename package: package path %q is the same as module path %q", meta.PkgPath, meta.Module.Path)
 	}
@@ -197,9 +215,10 @@ func prepareRenamePackageName(ctx context.Context, snapshot *cache.Snapshot, pgf
 	if err != nil {
 		return nil, err
 	}
+
 	return &PrepareItem{
 		Range: rng,
-		Text:  string(meta.Name),
+		Text:  string(meta.PkgPath),
 	}, nil
 }
 
@@ -209,8 +228,8 @@ func prepareRenamePackageName(ctx context.Context, snapshot *cache.Snapshot, pgf
 //
 // The resulting text is the signature of the function, which may be edited to
 // the new signature.
-func prepareRenameFuncSignature(pgf *parsego.File, pos token.Pos) (*PrepareItem, error) {
-	fdecl := funcKeywordDecl(pgf, pos)
+func prepareRenameFuncSignature(pgf *parsego.File, pos token.Pos, cursor inspector.Cursor) (*PrepareItem, error) {
+	fdecl := funcKeywordDecl(pos, cursor)
 	if fdecl == nil {
 		return nil, nil
 	}
@@ -233,17 +252,17 @@ func prepareRenameFuncSignature(pgf *parsego.File, pos token.Pos) (*PrepareItem,
 // nameBlankParams returns a copy of ftype with blank or unnamed params
 // assigned a unique name.
 func nameBlankParams(ftype *ast.FuncType) *ast.FuncType {
-	ftype = internalastutil.CloneNode(ftype)
+	ftype = astutil.CloneNode(ftype)
 
 	// First, collect existing names.
 	scope := make(map[string]bool)
-	for name := range goplsastutil.FlatFields(ftype.Params) {
+	for name := range astutil.FlatFields(ftype.Params) {
 		if name != nil {
 			scope[name.Name] = true
 		}
 	}
 	blanks := 0
-	for name, field := range goplsastutil.FlatFields(ftype.Params) {
+	for name, field := range astutil.FlatFields(ftype.Params) {
 		if name == nil {
 			name = ast.NewIdent("_")
 			field.Names = append(field.Names, name) // ok to append
@@ -264,17 +283,8 @@ func nameBlankParams(ftype *ast.FuncType) *ast.FuncType {
 
 // renameFuncSignature computes and applies the effective change signature
 // operation resulting from a 'renamed' (=rewritten) signature.
-func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, error) {
-	// Find the renamed signature.
-	pkg, pgf, err := NarrowestPackageForFile(ctx, snapshot, f.URI())
-	if err != nil {
-		return nil, err
-	}
-	pos, err := pgf.PositionPos(pp)
-	if err != nil {
-		return nil, err
-	}
-	fdecl := funcKeywordDecl(pgf, pos)
+func renameFuncSignature(ctx context.Context, pkg *cache.Package, pgf *parsego.File, pos token.Pos, snapshot *cache.Snapshot, cursor inspector.Cursor, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, error) {
+	fdecl := funcKeywordDecl(pos, cursor)
 	if fdecl == nil {
 		return nil, nil
 	}
@@ -295,11 +305,11 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 		return nil, fmt.Errorf("changing results not yet supported (got %d results, want %d)", got, want)
 	}
 	var resultTypes []string
-	for _, field := range goplsastutil.FlatFields(ftyp.Results) {
+	for _, field := range astutil.FlatFields(ftyp.Results) {
 		resultTypes = append(resultTypes, FormatNode(token.NewFileSet(), field.Type))
 	}
 	resultIndex := 0
-	for _, field := range goplsastutil.FlatFields(newType.Results) {
+	for _, field := range astutil.FlatFields(newType.Results) {
 		if FormatNode(token.NewFileSet(), field.Type) != resultTypes[resultIndex] {
 			return nil, fmt.Errorf("changing results not yet supported")
 		}
@@ -311,7 +321,7 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 		typ string
 	}
 	oldParams := make(map[string]paramInfo)
-	for name, field := range goplsastutil.FlatFields(ftyp.Params) {
+	for name, field := range astutil.FlatFields(ftyp.Params) {
 		oldParams[name.Name] = paramInfo{
 			idx: len(oldParams),
 			typ: types.ExprString(field.Type),
@@ -319,7 +329,7 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 	}
 
 	var newParams []int
-	for name, field := range goplsastutil.FlatFields(newType.Params) {
+	for name, field := range astutil.FlatFields(newType.Params) {
 		if name == nil {
 			return nil, fmt.Errorf("need named fields")
 		}
@@ -350,12 +360,8 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 
 // funcKeywordDecl returns the FuncDecl for which pos is in the 'func' keyword,
 // if any.
-func funcKeywordDecl(pgf *parsego.File, pos token.Pos) *ast.FuncDecl {
-	path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos)
-	if len(path) < 1 {
-		return nil
-	}
-	fdecl, _ := path[0].(*ast.FuncDecl)
+func funcKeywordDecl(pos token.Pos, cursor inspector.Cursor) *ast.FuncDecl {
+	fdecl, _ := cursorutil.FirstEnclosing[*ast.FuncDecl](cursor)
 	if fdecl == nil {
 		return nil
 	}
@@ -389,37 +395,78 @@ func checkRenamable(obj types.Object, node ast.Node) error {
 	return nil
 }
 
+// editsToDocChanges converts a map of uris to arrays of text edits to a list of document changes.
+func editsToDocChanges(ctx context.Context, snapshot *cache.Snapshot, edits map[protocol.DocumentURI][]protocol.TextEdit) ([]protocol.DocumentChange, error) {
+	var changes []protocol.DocumentChange
+	for uri, e := range edits {
+		fh, err := snapshot.ReadFile(ctx, uri)
+		if err != nil {
+			return nil, err
+		}
+		changes = append(changes, protocol.DocumentChangeEdit(fh, e))
+	}
+	return changes, nil
+}
+
 // Rename returns a map of TextEdits for each file modified when renaming a
-// given identifier within a package and a boolean value of true for renaming
-// package and false otherwise.
-func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, bool, error) {
+// given identifier within a package.
+func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) ([]protocol.DocumentChange, error) {
 	ctx, done := event.Start(ctx, "golang.Rename")
 	defer done()
 
-	if edits, err := renameFuncSignature(ctx, snapshot, f, pp, newName); err != nil {
-		return nil, false, err
-	} else if edits != nil {
-		return edits, false, nil
+	pkg, pgf, err := NarrowestPackageForFile(ctx, snapshot, f.URI())
+	if err != nil {
+		return nil, err
+	}
+	pos, err := pgf.PositionPos(pp)
+	if err != nil {
+		return nil, err
 	}
 
-	if !isValidIdentifier(newName) {
-		return nil, false, fmt.Errorf("invalid identifier to rename: %q", newName)
+	cur, ok := pgf.Cursor.FindByPos(pos, pos)
+	if !ok {
+		return nil, fmt.Errorf("can't find cursor for selection")
+	}
+
+	if edits, err := renameFuncSignature(ctx, pkg, pgf, pos, snapshot, cur, f, pp, newName); err != nil {
+		return nil, err
+	} else if edits != nil {
+		return editsToDocChanges(ctx, snapshot, edits)
 	}
 
 	// Cursor within package name declaration?
 	_, inPackageName, err := parsePackageNameDecl(ctx, snapshot, f, pp)
 	if err != nil {
-		return nil, false, err
+		return nil, err
 	}
 
-	var editMap map[protocol.DocumentURI][]diff.Edit
+	var (
+		editMap   map[protocol.DocumentURI][]diff.Edit
+		newPkgDir string // declared here so it can be used later to move the package's contents to the new directory
+	)
 	if inPackageName {
-		editMap, err = renamePackageName(ctx, snapshot, f, PackageName(newName))
+		countRenamePackage.Inc()
+		var (
+			newPkgName PackageName
+			newPkgPath PackagePath
+			err        error
+		)
+		moveSubpackages := snapshot.Options().RenameMovesSubpackages
+		if newPkgDir, newPkgName, newPkgPath, err = checkPackageRename(pkg, f, newName, moveSubpackages); err != nil {
+			return nil, err
+		}
+		editMap, err = renamePackage(ctx, snapshot, f, newPkgName, newPkgPath, newPkgDir, moveSubpackages)
+		if err != nil {
+			return nil, err
+		}
 	} else {
+		if !isValidIdentifier(newName) {
+			return nil, fmt.Errorf("invalid identifier to rename: %q", newName)
+		}
 		editMap, err = renameOrdinary(ctx, snapshot, f.URI(), pp, newName)
-	}
-	if err != nil {
-		return nil, false, err
+		if err != nil {
+			return nil, err
+		}
 	}
 
 	// Convert edits to protocol form.
@@ -448,21 +495,69 @@ func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp pro
 		// vendor/k8s.io/kubectl -> ../../staging/src/k8s.io/kubectl.
 		fh, err := snapshot.ReadFile(ctx, uri)
 		if err != nil {
-			return nil, false, err
+			return nil, err
 		}
 		data, err := fh.Content()
 		if err != nil {
-			return nil, false, err
+			return nil, err
 		}
 		m := protocol.NewMapper(uri, data)
 		textedits, err := protocol.EditsFromDiffEdits(m, edits)
 		if err != nil {
-			return nil, false, err
+			return nil, err
 		}
 		result[uri] = textedits
 	}
 
-	return result, inPackageName, nil
+	changes, err := editsToDocChanges(ctx, snapshot, result)
+	if err != nil {
+		return nil, err
+	}
+	if inPackageName {
+		oldDir := f.URI().DirPath()
+		if snapshot.Options().RenameMovesSubpackages {
+			// Update the last component of the file's enclosing directory.
+			changes = append(changes, protocol.DocumentChangeRename(
+				protocol.URIFromPath(oldDir),
+				protocol.URIFromPath(newPkgDir)))
+		} else {
+			// Not moving subpackages, so we need to move the files individually
+			// Move each file from oldDir to newPkgDir
+			files, err := os.ReadDir(oldDir)
+			if err != nil {
+				return nil, err
+			}
+			hasNestedDir := false
+			for _, f := range files {
+				if !f.IsDir() {
+					changes = append(changes,
+						protocol.DocumentChangeRename(
+							protocol.URIFromPath(filepath.Join(oldDir, f.Name())),
+							protocol.URIFromPath(filepath.Join(newPkgDir, f.Name()))),
+					)
+				} else {
+					hasNestedDir = true
+				}
+			}
+			// Delete oldPkgDir if it is now empty and newPkgDir is not a subdirectory of oldPkgDir.
+			// protocol.DeleteFile, when used with a directory, will only delete the directory
+			// if it is empty as long as the "recursive" option is set to its default value, false.
+			// TODO(mkalil): The above should be true according to the LSP spec but in VSCode
+			// we could get "Renamed failed to apply edits" errors when we include the edit
+			// to delete a directory that can't be deleted. For now, check if the oldDir has
+			// any nested directories before trying to delete it.
+			// TODO(mkalil): We could be deleting more directories. For example, if
+			// we have a package at a/b/c/d and we move it to a/z, it only deletes
+			// directory d and would leave directories b and c even if they are now
+			// empty.
+			oldPkgDir := f.URI().DirPath()
+			if !pathutil.InDir(oldPkgDir, newPkgDir) && !hasNestedDir {
+				changes = append(changes,
+					protocol.DocumentChangeDelete(protocol.URIFromPath(oldPkgDir)))
+			}
+		}
+	}
+	return changes, nil
 }
 
 // renameOrdinary renames an ordinary (non-package) name throughout the workspace.
@@ -474,11 +569,10 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, uri protocol.
 	// only package we need. (In case you're wondering why
 	// 'references' doesn't also want the widest variant: it
 	// computes the union across all variants.)
-	mps, err := snapshot.MetadataForFile(ctx, uri)
+	mps, err := snapshot.MetadataForFile(ctx, uri, true)
 	if err != nil {
 		return nil, err
 	}
-	metadata.RemoveIntermediateTestVariants(&mps)
 	if len(mps) == 0 {
 		return nil, fmt.Errorf("no package metadata for file %s", uri)
 	}
@@ -496,22 +590,30 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, uri protocol.
 	if err != nil {
 		return nil, err
 	}
-	var ok bool
-	cur, ok := pgf.Cursor.FindByPos(pos, pos) // of selected Ident or ImportSpec
-	if !ok {
-		return nil, fmt.Errorf("can't find cursor for selection")
-	}
-	targets, node, err := objectsAt(pkg.TypesInfo(), pgf.File, pos)
+	cur, _ := pgf.Cursor.FindByPos(pos, pos) // cannot fail
+
+	targets, err := objectsAt(pkg.TypesInfo(), cur)
 	if err != nil {
-		return nil, err
+		// Check if we are renaming an ident inside its doc comment. The call to
+		// objectsAt will have returned an error in this case.
+		// TODO(adonovan): move this logic into objectsAt.
+		var ok bool
+		cur, ok = docCommentPosToIdent(pgf, pos, cur)
+		if !ok {
+			return nil, err
+		}
+		id := cur.Node().(*ast.Ident)
+		obj := pkg.TypesInfo().Defs[id]
+		if obj == nil {
+			return nil, fmt.Errorf("error fetching types.Object for ident %q", id.Name)
+		}
+		// Change rename target to the ident.
+		targets = []objectAt{{obj, cur}}
 	}
 
 	// Pick a representative object arbitrarily.
 	// (All share the same name, pos, and kind.)
-	var obj types.Object
-	for obj = range targets {
-		break
-	}
+	obj, node := targets[0].obj, targets[0].cur.Node()
 	if obj.Name() == newName {
 		return nil, fmt.Errorf("old and new names are the same: %s", newName)
 	}
@@ -577,8 +679,8 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, uri protocol.
 	// Nonexported? Search locally.
 	if declObjPath == "" {
 		var objects []types.Object
-		for obj := range targets {
-			objects = append(objects, obj)
+		for _, o := range targets {
+			objects = append(objects, o.obj)
 		}
 
 		editMap, _, err := renameObjects(newName, pkg, objects...)
@@ -595,8 +697,8 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, uri protocol.
 		// intent for the broader renaming; renaming a use of
 		// the receiver effects only the local renaming.
 		if id, ok := cur.Node().(*ast.Ident); ok && id.Pos() == obj.Pos() {
-			if curDecl, ok := moreiters.First(cur.Enclosing((*ast.FuncDecl)(nil))); ok {
-				decl := curDecl.Node().(*ast.FuncDecl) // enclosing func
+			// enclosing func
+			if decl, _ := cursorutil.FirstEnclosing[*ast.FuncDecl](cur); decl != nil {
 				if decl.Recv != nil &&
 					len(decl.Recv.List) > 0 &&
 					len(decl.Recv.List[0].Names) > 0 {
@@ -714,7 +816,7 @@ func renameReceivers(pkg *cache.Package, recv *types.Var, newName string, editMa
 // selectors used only in an ITV, but life is short. Also sin must be
 // punished.)
 func typeCheckReverseDependencies(ctx context.Context, snapshot *cache.Snapshot, declURI protocol.DocumentURI, transitive bool) ([]*cache.Package, error) {
-	variants, err := snapshot.MetadataForFile(ctx, declURI)
+	variants, err := snapshot.MetadataForFile(ctx, declURI, false)
 	if err != nil {
 		return nil, err
 	}
@@ -847,33 +949,39 @@ func renameExported(pkgs []*cache.Package, declPkgPath PackagePath, declObjPath
 	return allEdits, nil
 }
 
-// renamePackageName renames package declarations, imports, and go.mod files.
-func renamePackageName(ctx context.Context, s *cache.Snapshot, f file.Handle, newName PackageName) (map[protocol.DocumentURI][]diff.Edit, error) {
+// renamePackage renames package declarations, imports, and go.mod files.
+//
+// f is the file originating the rename, and therefore f.URI().Dir() is the
+// current package directory. newName, newPath, and newDir describe the renaming.
+func renamePackage(ctx context.Context, s *cache.Snapshot, f file.Handle, newName PackageName, newPath PackagePath, newDir string, moveSubpackages bool) (map[protocol.DocumentURI][]diff.Edit, error) {
 	// Rename the package decl and all imports.
-	renamingEdits, err := renamePackage(ctx, s, f, newName)
+	renamingEdits := make(map[protocol.DocumentURI][]diff.Edit)
+	err := updatePackageDeclsAndImports(ctx, s, f, newName, newPath, renamingEdits, moveSubpackages)
 	if err != nil {
 		return nil, err
 	}
 
-	// Update the last component of the file's enclosing directory.
-	oldBase := f.URI().DirPath()
-	newPkgDir := filepath.Join(filepath.Dir(oldBase), string(newName))
+	err = updateModFiles(ctx, s, f.URI().DirPath(), newDir, renamingEdits, moveSubpackages)
+	if err != nil {
+		return nil, err
+	}
 
-	// Update any affected replace directives in go.mod files.
-	// TODO(adonovan): extract into its own function.
-	//
-	// Get all workspace modules.
-	// TODO(adonovan): should this operate on all go.mod files,
-	// irrespective of whether they are included in the workspace?
+	return renamingEdits, nil
+}
+
+// Update any affected replace directives in go.mod files.
+// TODO(adonovan): should this operate on all go.mod files,
+// irrespective of whether they are included in the workspace?
+func updateModFiles(ctx context.Context, s *cache.Snapshot, oldDir string, newPkgDir string, renamingEdits map[protocol.DocumentURI][]diff.Edit, moveSubpackages bool) error {
 	modFiles := s.View().ModFiles()
 	for _, m := range modFiles {
 		fh, err := s.ReadFile(ctx, m)
 		if err != nil {
-			return nil, err
+			return err
 		}
 		pm, err := s.ParseMod(ctx, fh)
 		if err != nil {
-			return nil, err
+			return err
 		}
 
 		modFileDir := pm.URI.DirPath()
@@ -884,15 +992,16 @@ func renamePackageName(ctx context.Context, s *cache.Snapshot, f file.Handle, ne
 			if !strings.HasPrefix(r.New.Path, "/") && !strings.HasPrefix(r.New.Path, "./") && !strings.HasPrefix(r.New.Path, "../") {
 				continue
 			}
-
-			replacedPath := r.New.Path
+			replacedDir := r.New.Path
 			if strings.HasPrefix(r.New.Path, "./") || strings.HasPrefix(r.New.Path, "../") {
-				replacedPath = filepath.Join(modFileDir, r.New.Path)
+				replacedDir = filepath.Join(modFileDir, r.New.Path)
 			}
 
 			// TODO: Is there a risk of converting a '\' delimited replacement to a '/' delimited replacement?
-			if !strings.HasPrefix(filepath.ToSlash(replacedPath)+"/", filepath.ToSlash(oldBase)+"/") {
-				continue // not affected by the package renaming
+
+			if moveSubpackages && !strings.HasPrefix(filepath.ToSlash(replacedDir)+"/", filepath.ToSlash(oldDir)+"/") ||
+				!moveSubpackages && !(filepath.ToSlash(replacedDir) == filepath.ToSlash(oldDir)) {
+				continue //not affected by the package renanming
 			}
 
 			affectedReplaces = append(affectedReplaces, r)
@@ -903,20 +1012,20 @@ func renamePackageName(ctx context.Context, s *cache.Snapshot, f file.Handle, ne
 		}
 		copied, err := modfile.Parse("", pm.Mapper.Content, nil)
 		if err != nil {
-			return nil, err
+			return err
 		}
 
 		for _, r := range affectedReplaces {
-			replacedPath := r.New.Path
+			replacedDir := r.New.Path
 			if strings.HasPrefix(r.New.Path, "./") || strings.HasPrefix(r.New.Path, "../") {
-				replacedPath = filepath.Join(modFileDir, r.New.Path)
+				replacedDir = filepath.Join(modFileDir, r.New.Path)
 			}
 
-			suffix := strings.TrimPrefix(replacedPath, oldBase)
+			suffix := strings.TrimPrefix(replacedDir, oldDir)
 
 			newReplacedPath, err := filepath.Rel(modFileDir, newPkgDir+suffix)
 			if err != nil {
-				return nil, err
+				return err
 			}
 
 			newReplacedPath = filepath.ToSlash(newReplacedPath)
@@ -926,72 +1035,67 @@ func renamePackageName(ctx context.Context, s *cache.Snapshot, f file.Handle, ne
 			}
 
 			if err := copied.AddReplace(r.Old.Path, "", newReplacedPath, ""); err != nil {
-				return nil, err
+				return err
 			}
 		}
 
 		copied.Cleanup()
 		newContent, err := copied.Format()
 		if err != nil {
-			return nil, err
+			return err
 		}
 
 		// Calculate the edits to be made due to the change.
 		edits := diff.Bytes(pm.Mapper.Content, newContent)
 		renamingEdits[pm.URI] = append(renamingEdits[pm.URI], edits...)
 	}
-
-	return renamingEdits, nil
+	return nil
 }
 
-// renamePackage computes all workspace edits required to rename the package
+// updatePackageDeclsAndImports computes all workspace edits required to rename the package
 // described by the given metadata, to newName, by renaming its package
 // directory.
 //
 // It updates package clauses and import paths for the renamed package as well
 // as any other packages affected by the directory renaming among all packages
 // known to the snapshot.
-func renamePackage(ctx context.Context, s *cache.Snapshot, f file.Handle, newName PackageName) (map[protocol.DocumentURI][]diff.Edit, error) {
+func updatePackageDeclsAndImports(ctx context.Context, s *cache.Snapshot, f file.Handle, newName PackageName, newPkgPath PackagePath, renamingEdits map[protocol.DocumentURI][]diff.Edit, moveSubpackages bool) error {
 	if strings.HasSuffix(string(newName), "_test") {
-		return nil, fmt.Errorf("cannot rename to _test package")
+		return fmt.Errorf("cannot rename to _test package")
 	}
 
 	// We need metadata for the relevant package and module paths.
 	// These should be the same for all packages containing the file.
 	meta, err := s.NarrowestMetadataForFile(ctx, f.URI())
 	if err != nil {
-		return nil, err
+		return err
 	}
 
 	oldPkgPath := meta.PkgPath
 	if meta.Module == nil {
-		return nil, fmt.Errorf("cannot rename package: missing module information for package %q", meta.PkgPath)
+		return fmt.Errorf("cannot rename package: missing module information for package %q", meta.PkgPath)
 	}
 	modulePath := PackagePath(meta.Module.Path)
 	if modulePath == oldPkgPath {
-		return nil, fmt.Errorf("cannot rename package: module path %q is the same as the package path, so renaming the package directory would have no effect", modulePath)
+		return fmt.Errorf("cannot rename package: module path %q is the same as the package path, so renaming the package directory would have no effect", modulePath)
 	}
 
-	newPathPrefix := path.Join(path.Dir(string(oldPkgPath)), string(newName))
-
 	// We must inspect all packages, not just direct importers,
-	// because we also rename subpackages, which may be unrelated.
+	// because we might also rename subpackages, which may be unrelated.
 	// (If the renamed package imports a subpackage it may require
 	// edits to both its package and import decls.)
 	allMetadata, err := s.AllMetadata(ctx)
 	if err != nil {
-		return nil, err
+		return err
 	}
-
 	// Rename package and import declarations in all relevant packages.
-	edits := make(map[protocol.DocumentURI][]diff.Edit)
 	for _, mp := range allMetadata {
 		// Special case: x_test packages for the renamed package will not have the
 		// package path as a dir prefix, but still need their package clauses
 		// renamed.
 		if mp.PkgPath == oldPkgPath+"_test" {
-			if err := renamePackageClause(ctx, mp, s, newName+"_test", edits); err != nil {
-				return nil, err
+			if err := renamePackageClause(ctx, mp, s, newName+"_test", renamingEdits); err != nil {
+				return err
 			}
 			continue
 		}
@@ -999,13 +1103,14 @@ func renamePackage(ctx context.Context, s *cache.Snapshot, f file.Handle, newNam
 		// Subtle: check this condition before checking for valid module info
 		// below, because we should not fail this operation if unrelated packages
 		// lack module info.
-		if !strings.HasPrefix(string(mp.PkgPath)+"/", string(oldPkgPath)+"/") {
+		if moveSubpackages && !pathutil.InDir(filepath.FromSlash(string(oldPkgPath)), filepath.FromSlash(string(mp.PkgPath))) ||
+			!moveSubpackages && mp.PkgPath != oldPkgPath {
 			continue // not affected by the package renaming
 		}
 
 		if mp.Module == nil {
 			// This check will always fail under Bazel.
-			return nil, fmt.Errorf("cannot rename package: missing module information for package %q", mp.PkgPath)
+			return fmt.Errorf("cannot rename package: missing module information for package %q", mp.PkgPath)
 		}
 
 		if modulePath != PackagePath(mp.Module.Path) {
@@ -1014,24 +1119,23 @@ func renamePackage(ctx context.Context, s *cache.Snapshot, f file.Handle, newNam
 
 		// Renaming a package consists of changing its import path and package name.
 		suffix := strings.TrimPrefix(string(mp.PkgPath), string(oldPkgPath))
-		newPath := newPathPrefix + suffix
+		newImportPath := string(newPkgPath) + suffix
 
 		pkgName := mp.Name
 		if mp.PkgPath == oldPkgPath {
 			pkgName = newName
 
-			if err := renamePackageClause(ctx, mp, s, newName, edits); err != nil {
-				return nil, err
+			if err := renamePackageClause(ctx, mp, s, newName, renamingEdits); err != nil {
+				return err
 			}
 		}
-
-		imp := ImportPath(newPath) // TODO(adonovan): what if newPath has vendor/ prefix?
-		if err := renameImports(ctx, s, mp, imp, pkgName, edits); err != nil {
-			return nil, err
+		imp := ImportPath(newImportPath) // TODO(adonovan): what if newImportPath has vendor/ prefix?
+		if err := renameImports(ctx, s, mp, imp, pkgName, renamingEdits); err != nil {
+			return err
 		}
 	}
 
-	return edits, nil
+	return nil
 }
 
 // renamePackageClause computes edits renaming the package clause of files in
@@ -1098,6 +1202,16 @@ func renameImports(ctx context.Context, snapshot *cache.Snapshot, mp *metadata.P
 					continue // not the import we're looking for
 				}
 
+				// If we are moving a package from a non-internal directory to
+				// an internal directory, and there is an importer located in a
+				// different module, then it is an invalid import.
+				//
+				// We do not prevent moves from an internal directory to a
+				// non-internal directory.
+				if !metadata.IsValidImport(rdep.PkgPath, metadata.PackagePath(newPath), snapshot.View().Type() != cache.GoPackagesDriverView) {
+					return fmt.Errorf("invalid: package move would result in illegal internal import")
+				}
+
 				// If the import does not explicitly specify
 				// a local name, then we need to invoke the
 				// type checker to locate references to update.
@@ -1328,7 +1442,7 @@ func (r *renamer) update() (map[protocol.DocumentURI][]diff.Edit, error) {
 		// the case clause to the switch ident.
 		// This may result in duplicate edits, but we de-dup later.
 		if _, ok := item.node.(*ast.CaseClause); ok {
-			path, _ := astutil.PathEnclosingInterval(pgf.File, item.obj.Pos(), item.obj.Pos())
+			path, _ := goastutil.PathEnclosingInterval(pgf.File, item.obj.Pos(), item.obj.Pos())
 			item.node = path[0].(*ast.Ident)
 		}
 
@@ -1344,7 +1458,8 @@ func (r *renamer) update() (map[protocol.DocumentURI][]diff.Edit, error) {
 			continue
 		}
 
-		doc := docComment(pgf, item.node.(*ast.Ident))
+		cur, _ := pgf.Cursor.FindNode(item.node) // can't fail
+		doc := docComment(pgf, cur)
 		if doc == nil {
 			continue
 		}
@@ -1584,10 +1699,11 @@ func (r *docLinkRenamer) update(pgf *parsego.File) (result []diff.Edit, err erro
 }
 
 // docComment returns the doc for an identifier within the specified file.
-func docComment(pgf *parsego.File, id *ast.Ident) *ast.CommentGroup {
-	nodes, _ := astutil.PathEnclosingInterval(pgf.File, id.Pos(), id.End())
-	for _, node := range nodes {
-		switch decl := node.(type) {
+func docComment(pgf *parsego.File, curId inspector.Cursor) *ast.CommentGroup {
+	// (Strictly it needn't be an identifier; only its Pos is used.)
+	id := curId.Node().(*ast.Ident)
+	for cur := range curId.Enclosing() {
+		switch decl := cur.Node().(type) {
 		case *ast.FuncDecl:
 			return decl.Doc
 		case *ast.Field:
@@ -1615,7 +1731,7 @@ func docComment(pgf *parsego.File, id *ast.Ident) *ast.CommentGroup {
 			}
 
 			identLine := safetoken.Line(pgf.Tok, id.Pos())
-			for _, comment := range nodes[len(nodes)-1].(*ast.File).Comments {
+			for _, comment := range pgf.File.Comments {
 				if comment.Pos() > id.Pos() {
 					// Comment is after the identifier.
 					continue
@@ -1633,11 +1749,47 @@ func docComment(pgf *parsego.File, id *ast.Ident) *ast.CommentGroup {
 	return nil
 }
 
+// docCommentPosToIdent returns a cursor for the identifier whose doc
+// comment contains pos, if any. The pos must be within an occurrence
+// of the identifier's name, otherwise it returns zero.
+func docCommentPosToIdent(pgf *parsego.File, pos token.Pos, cur inspector.Cursor) (inspector.Cursor, bool) {
+	for curId := range cur.Preorder((*ast.Ident)(nil)) {
+		id := curId.Node().(*ast.Ident)
+		if pos > id.Pos() {
+			continue // Doc comments are not located after an ident.
+		}
+		doc := docComment(pgf, curId)
+		if doc == nil || !(doc.Pos() <= pos && pos < doc.End()) {
+			continue
+		}
+
+		docRegexp := regexp.MustCompile(`\b` + id.Name + `\b`)
+		for _, comment := range doc.List {
+			if isDirective(comment.Text) || !(comment.Pos() <= pos && pos < comment.End()) {
+				continue
+			}
+			start := comment.Pos()
+			text, err := pgf.NodeText(comment)
+			if err != nil {
+				return inspector.Cursor{}, false
+			}
+			for _, locs := range docRegexp.FindAllIndex(text, -1) {
+				matchStart := start + token.Pos(locs[0])
+				matchEnd := start + token.Pos(locs[1])
+				if matchStart <= pos && pos <= matchEnd {
+					return curId, true
+				}
+			}
+		}
+	}
+	return inspector.Cursor{}, false
+}
+
 // updatePkgName returns the updates to rename a pkgName in the import spec by
 // only modifying the package name portion of the import declaration.
 func (r *renamer) updatePkgName(pgf *parsego.File, pkgName *types.PkgName) (diff.Edit, error) {
 	// Modify ImportSpec syntax to add or remove the Name as needed.
-	path, _ := astutil.PathEnclosingInterval(pgf.File, pkgName.Pos(), pkgName.Pos())
+	path, _ := goastutil.PathEnclosingInterval(pgf.File, pkgName.Pos(), pkgName.Pos())
 	if len(path) < 2 {
 		return diff.Edit{}, fmt.Errorf("no path enclosing interval for %s", pkgName.Name())
 	}
@@ -1670,7 +1822,7 @@ func parsePackageNameDecl(ctx context.Context, snapshot *cache.Snapshot, fh file
 	// Careful: because we used parsego.Header,
 	// pgf.Pos(ppos) may be beyond EOF => (0, err).
 	pos, _ := pgf.PositionPos(ppos)
-	return pgf, goplsastutil.NodeContains(pgf.File.Name, pos), nil
+	return pgf, astutil.NodeContainsPos(pgf.File.Name, pos), nil
 }
 
 // posEdit returns an edit to replace the (start, end) range of tf with 'new'.
